import os
import typing
import importlib


class Singleton(type):
    def __init__(
        cls,
        name: str,
        bases: typing.Iterable[str],
        namespace: typing.Dict[str, typing.Any],
    ) -> None:
        cls.instance = None

    def __call__(cls, *args, **kwargs) -> typing.Any:
        if cls.instance is None:
            cls.instance = super().__call__(*args, **kwargs)
        return cls.instance


def _import_module(name: str) -> None:
    """
    try importlib.import_module, nothing to do when module not be found.
    """
    from .config import config

    if os.path.exists(os.path.join(config.path, name + ".py")) or os.path.exists(
        os.path.join(config.path, name, "__init__.py")
    ):
        importlib.import_module(name)



import os
import re
import copy
import typing
import asyncio
from abc import abstractmethod, ABCMeta

from aiofiles import open as asyncopen
from starlette.datastructures import UploadFile
from index.config import config

from .utils import merge_mapping


EMPTY_VALUE = ("", {}, [], None, ())

ALLOWS_TYPE = ("array", "boolean", "integer", "number", "object", "string")


class VerifyError(Exception):
    pass


class FieldVerifyError(VerifyError):
    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class ModelFieldVerifyError(FieldVerifyError):
    def __init__(self, message: typing.Dict[str, typing.Any]) -> None:
        super().__init__(message)
        self.message = message


class Field(metaclass=ABCMeta):
    def __init__(
        self,
        *,
        description: str = "",
        example: typing.Any = None,
        allow_null: bool = False,
        default: typing.Any = None,
    ):
        self.description = description
        self.example = example
        self.allow_null = allow_null
        self.default = default

    def check_null(self, value: typing.Any) -> typing.Any:
        if value is None:
            if callable(self.default):
                value = self.default()
            value = copy.deepcopy(self.default)

        if value in EMPTY_VALUE and not self.allow_null:
            raise FieldVerifyError("Not allowed to be empty.")
        return value

    @abstractmethod
    async def verify(self, value: typing.Any) -> typing.Any:
        """Verify and parse data into Python objects"""

    @abstractmethod
    async def json(self, value: typing.Any) -> typing.Any:
        """parse data from Python objects"""

    def openapi(self) -> typing.Dict[str, typing.Any]:
        schema = {
            "description": self.description,
        }
        if self.default not in EMPTY_VALUE:
            schema["default"] = self.default
        return schema


class MatchFieldMeta(type):
    def __init__(
        self,
        name: str,
        bases: typing.Iterable[typing.Any],
        namespace: typing.Dict[str, typing.Any],
    ):
        fields = {}
        for key, value in namespace.items():
            if isinstance(value, Field):
                fields[key] = value
        self.fields = fields
        super().__init__(name, bases, namespace)


class BaseModel(metaclass=MatchFieldMeta):
    fields: typing.Dict[str, Field]


class ChoiceField(Field, metaclass=ABCMeta):
    def __init__(self, *, choices: typing.Iterable[str] = (), **kwargs):
        super().__init__(**kwargs)
        self.choices = choices
        if choices and self.default is not None:
            assert self.default in choices, "default value must be in choices"

    def check_choice(self, value: typing.Any) -> typing.Any:
        if self.choices and value not in self.choices:
            raise FieldVerifyError(f"value must be in {self.choices}")
        return value

    @abstractmethod
    async def verify(self, value: typing.Any) -> typing.Any:
        pass

    @abstractmethod
    async def json(self, value: typing.Any) -> typing.Any:
        pass

    def openapi(self) -> typing.Dict[str, typing.Any]:
        schema = super().openapi()
        if self.choices:
            schema["enum"] = list(self.choices)
        return schema


class FileField(Field):
    def __init__(
        self,
        prefix: str,
        *,
        description: str = "",
        allow_null: bool = False,
        save: typing.Callable[[UploadFile], typing.Awaitable[str]] = None,
    ):
        super().__init__(description=description, allow_null=allow_null)
        self.prefix = prefix
        assert not prefix.startswith("/"), 'Prefix cannot start with "/".'
        if save is not None:
            setattr(self, "save", save)

    async def save(self, file: UploadFile) -> str:
        abspath = os.path.join(config.path, self.prefix, file.filename)
        if os.path.exists(abspath):
            raise FileExistsError(abspath)
        async with asyncopen(abspath, "w+") as target:
            while True:
                data = await file.read(1024)
                if not data:
                    break
                await target.write(data)
        return abspath

    async def verify(self, value: UploadFile) -> str:
        """return file path/url"""
        try:
            return await self.save(value)
        except FileExistsError:
            raise FieldVerifyError(f'This file "{value.filename}" already exists.')

    def json(self, value: typing.Any) -> typing.NoReturn:
        raise NotImplementedError("FileField not implement json")

    def openapi(self) -> typing.Dict[str, typing.Any]:
        schema = {"type": "string", "format": "binary"}
        if self.description:
            schema.update({"description": self.description})
        return schema


class BooleanField(Field):
    def __init__(self, *, default: typing.Any = None, **kwargs):
        assert default is None or default in (
            True,
            False,
        ), "default in BooleanField must be bool"
        kwargs["default"] = default
        super().__init__(**kwargs)

    def verify(self, value: typing.Any) -> bool:
        value = self.check_null(value)
        return value != "0"

    def json(self, value: typing.Any) -> typing.Any:
        return bool(value)

    def openapi(self) -> typing.Dict[str, typing.Any]:
        schema = super().openapi()
        schema.update({"type": "boolean"})
        return schema


class IntField(ChoiceField):
    def verify(self, value: typing.Any) -> int:
        value = self.check_null(value)
        value = self.check_choice(value)
        try:
            return int(value)
        except ValueError:
            raise FieldVerifyError("Must be an integer.")

    def json(self, value: typing.Any) -> typing.Any:
        return int(value)

    def openapi(self) -> typing.Dict[str, typing.Any]:
        schema = super().openapi()
        schema.update({"type": "integer"})
        return schema


class FloatField(ChoiceField):
    def verify(self, value: typing.Any) -> float:
        value = self.check_null(value)
        value = self.check_choice(value)
        try:
            return float(value)
        except ValueError:
            raise FieldVerifyError("Must be an floating point number.")

    def json(self, value: typing.Any) -> typing.Any:
        return float(value)

    def openapi(self) -> typing.Dict[str, typing.Any]:
        schema = super().openapi()
        schema.update({"type": "number"})
        return schema


class StrField(ChoiceField):
    def verify(self, value: typing.Any) -> str:
        value = self.check_null(value)
        value = self.check_choice(value)
        return str(value)

    def json(self, value: typing.Any) -> typing.Any:
        return str(value)

    def openapi(self) -> typing.Dict[str, typing.Any]:
        schema = super().openapi()
        schema.update({"type": "string"})
        return schema


class EmailField(StrField):
    EMAIL_FORMAT = re.compile(r"(?P<name>\S+?)@(?P<domain>\S+?\.\S+?)")

    def check_format(self, value: str) -> str:
        if self.EMAIL_FORMAT.match(value) is None:
            raise FieldVerifyError("Must be a string in email format.")
        return value

    def verify(self, value: typing.Any) -> str:
        value = super().verify(value)
        return self.check_format(value)

    def openapi(self) -> typing.Dict[str, typing.Any]:
        schema = super().openapi()
        schema.update({"format": "email"})
        return schema


class ListField(Field):
    def __init__(self, field: Field, **kwargs):
        if "default" not in kwargs:
            kwargs["default"] = []
        assert isinstance(
            kwargs["default"], list
        ), "default in ListField must be `list`"
        super().__init__(**kwargs)
        self.field = field

    async def verify(self, value: typing.Iterable) -> typing.List[typing.Any]:
        result = []
        value = self.check_null(value)
        for each in value:
            data = self.field.verify(each)
            if asyncio.iscoroutine(data):
                data = await data
            result.append(data)
        return result

    def json(self, value: typing.Iterable) -> typing.List:
        return [self.field.json(item) for item in value]

    def openapi(self) -> typing.Dict[str, typing.Any]:
        schema = super().openapi()
        schema.update({"type": "array", "items": self.field.openapi()})
        return schema


class Model(BaseModel):
    content_type = "application/json"

    def __init__(
        self,
        raw_data: typing.Dict[str, typing.Any],
        *,
        default: typing.Dict[str, typing.Any] = None,
    ) -> None:
        if default:
            self.raw_data = merge_mapping(raw_data, default)
        else:
            self.raw_data = raw_data
        self.data: typing.Dict[str, typing.Any] = {}

    async def clean(self) -> typing.Dict[str, str]:
        errors = {}
        for name, field in self.fields.items():
            try:
                data = field.verify(self.raw_data.get(name))
                if asyncio.iscoroutine(data):
                    data = await data
                self.data[name] = data
                setattr(self, name, data)
            except FieldVerifyError as e:
                errors[name] = e.message
        return errors

    @classmethod
    async def serialization(
        cls, value: typing.Dict[str, typing.Any]
    ) -> typing.Dict[str, typing.Any]:
        result = {}
        for name, field in cls.fields.items():
            data = field.json(value.get(name))
            if asyncio.iscoroutine(data):
                data = await data
            result[name] = data
        return result


class ModelField(Field):
    def __init__(
        self, model: BaseModel, *, default: typing.Dict[str, typing.Any] = {}, **kwargs
    ):
        kwargs["default"] = None
        super().__init__(**kwargs)
        self.model = model
        self.default_model_value = default

    async def verify(self, value: typing.Any) -> typing.Any:
        model = self.model(value, default=self.default_model_value)
        errors = await model.clean()
        if errors:
            raise ModelFieldVerifyError(errors)
        return model

    async def json(self, value: typing.Any) -> typing.Any:
        return await self.model.serialization(value)

    def openapi(self) -> typing.Dict[str, typing.Any]:
        schema = super().openapi()
        schema.update(
            {
                "type": "object",
                "properties": {
                    name: field.openapi() for name, field in self.model.fields.items()
                },
            }
        )
        return schema



"""
Maybe more repsonse type will be done in the future
"""
import typing
import functools

import yaml
from starlette.background import BackgroundTask
from starlette.responses import (
    Response,
    HTMLResponse,
    PlainTextResponse,
    JSONResponse,
    RedirectResponse,
    StreamingResponse,
    FileResponse,
)
from starlette.templating import Jinja2Templates

from .types import typeassert

__all__ = [
    "register_type",
    "Response",
    "HTMLResponse",
    "PlainTextResponse",
    "JSONResponse",
    "YAMLResponse",
    "RedirectResponse",
    "StreamingResponse",
    "FileResponse",
    "TemplateResponse",
]


TemplateResponse = Jinja2Templates(directory="templates").TemplateResponse


class YAMLResponse(Response):
    media_type = "text/yaml"

    def render(self, content: typing.Any) -> bytes:
        return yaml.dump(content, indent=2).encode("utf8")


class AutoResponseType:

    type_map = {}

    @classmethod
    def register_type(cls, type_: typing.Any) -> typing.Callable:
        def register_func(func: typing.Callable) -> typing.Callable:
            cls.type_map[type_] = func

            def wrapper(*args, **kwargs) -> Response:
                return func(*args, **kwargs)

            return wrapper

        return register_func

    @classmethod
    def automatic(cls, *args) -> Response:

        # Response or Response subclass
        if isinstance(args[0], Response):
            return args[0]

        try:
            return cls.type_map[type(args[0])](*args)
        except KeyError:
            raise TypeError(
                f"Cannot find automatic handler for this type: {type(args[0])}"
            )


register_type = functools.partial(AutoResponseType.register_type)
automatic = functools.partial(AutoResponseType.automatic)


@register_type(dict)
@typeassert
def json_type(
    body: dict,
    status: int = 200,
    headers: dict = None,
    background: BackgroundTask = None,
) -> Response:

    return JSONResponse(body, status, headers, background=background)


@register_type(str)
@typeassert
def text_type(
    body: str,
    status: int = 200,
    headers: dict = None,
    background: BackgroundTask = None,
) -> Response:

    return PlainTextResponse(body, status, headers, background=background)



import typing

from starlette.requests import Request
from starlette.responses import Response

from .types import AsyncCallable
from .concurrency import keepasync


class MiddlewareMixin(metaclass=keepasync("process_request", "process_response")):

    ChildMiddlewares: typing.Iterable[typing.Callable] = ()

    def __init__(self, get_response: AsyncCallable) -> None:
        self.get_response = self.mount_middleware(get_response)

    def mount_middleware(self, get_response: AsyncCallable) -> AsyncCallable:
        for base_middleware in self.ChildMiddlewares:
            get_response = base_middleware(get_response)
        return get_response

    async def __call__(self, request: Request) -> Response:
        response = None
        if hasattr(self, "process_request"):
            response = await self.process_request(request)

        if response is None:
            response = await self.get_response(request)

        if hasattr(self, "process_response"):
            response = await self.process_response(request, response)
        return response



import os
import importlib
from copy import deepcopy
from inspect import signature

from starlette.types import Scope, Receive, Send
from starlette.endpoints import Request, Response
from starlette.exceptions import HTTPException

from index.responses import (
    JSONResponse,
    YAMLResponse,
    HTMLResponse,
)
from index.config import config

from .models import Model
from .functions import Schema


def get_views():
    views_path = os.path.join(config.path, "views").replace("\\", "/")
    for root, _, files in os.walk(views_path):
        for file in files:
            if not file.endswith(".py"):
                continue
            if file == "__init__.py":
                continue
            abspath = os.path.join(root, file)
            relpath = os.path.relpath(abspath, config.path).replace("\\", "/")
            module_name = relpath[:-3].replace("/", ".")
            module = importlib.import_module(module_name)
            path = relpath[len("views") : -3]
            if path.endswith("index"):
                path = path[: -len("index")]
            if abspath.startswith(views_path):
                yield path, module


class OpenAPI:
    def __init__(
        self, title: str, description: str, version: str, *, media_type="yaml"
    ):
        """
        media_type: yaml or json
        """
        assert media_type in ("yaml", "json"), "media_type must in 'yaml' or 'json'"

        info = {"title": title, "description": description, "version": version}
        self.openapi = {"openapi": "3.0.0", "info": info, "paths": {}}
        self.media_type = media_type

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http" or scope["path"] not in ("/", "/get"):
            raise HTTPException(404)
        request = Request(scope, receive, send)

        if scope["path"] == "/get":
            handler = getattr(self, "docs")
        elif scope["path"] == "/":
            handler = getattr(self, "template")
        response = await handler(request)
        await response(scope, receive, send)

    async def template(self, request: Request) -> Response:
        return HTMLResponse(DEFAULT_TEMPLATE)

    async def docs(self, request: Request) -> Response:
        openapi = deepcopy(self.openapi)
        openapi["servers"] = [
            {
                "url": f"{request.url.scheme}://{request.url.netloc}",
                "description": "Current server",
            }
        ]

        paths = openapi["paths"]
        for path, view in get_views():
            viewclass = view.HTTP()
            paths[path] = {}
            for method in viewclass.allowed_methods():
                if method == "OPTIONS":
                    continue
                method = method.lower()

                sig = signature(getattr(viewclass, method))
                paths[path][method] = {}

                doc = getattr(viewclass, method).__doc__
                if isinstance(doc, str):
                    doc = doc.strip()
                    paths[path][method].update(
                        {
                            "summary": doc.splitlines()[0],
                            "description": "\n".join(doc.splitlines()[1:]).strip(),
                        }
                    )

                query = sig.parameters.get("query")
                if query and issubclass(query.annotation, Model):
                    paths[path][method]["parameters"] = Schema.in_query(
                        query.annotation
                    )

                body = sig.parameters.get("body")
                if body and issubclass(body.annotation, Model):
                    paths[path][method]["requestBody"] = {
                        "required": True,
                        "content": {
                            body.annotation.content_type: {
                                "schema": Schema.request_body(body.annotation)
                            }
                        },
                    }
                    description = body.annotation.__doc__
                    if description:
                        paths[path][method]["requestBody"]["description"] = description

                try:
                    resps = getattr(getattr(viewclass, method), "__resps__")
                except AttributeError:
                    pass
                else:
                    repsonses = paths[path][method]["responses"] = {}
                    for status, content in resps.items():
                        repsonses[status] = {
                            "description": content["description"],
                        }
                        if content["model"] is not None:
                            repsonses[status]["content"] = {
                                content["model"].content_type: {
                                    "schema": Schema.response(content["model"])
                                }
                            }

                if not paths[path][method]:
                    del paths[path][method]

            if not paths[path]:
                del paths[path]

        if self.media_type == "yaml" or request.query_params.get("type") == "yaml":
            return YAMLResponse(openapi)

        if self.media_type == "json" or request.query_params.get("type") == "json":
            return JSONResponse(openapi)


DEFAULT_TEMPLATE = """
<!DOCTYPE html>
<html>
  <head>
    <title>OpenAPI power by Index.py</title>
    <!-- needed for adaptive design -->
    <meta charset="utf-8"/>
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <link href="https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700" rel="stylesheet">

    <style>
      body {
        margin: 0;
        padding: 0;
      }
    </style>
  </head>
  <body>
    <redoc spec-url='get'></redoc>
    <script src="https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js"> </script>
  </body>
</html>
"""



import typing
from inspect import signature
from functools import wraps

from .config import Config


AsyncCallable = typing.Callable[[typing.Any], typing.Awaitable]

# WSGI: view PEP3333
Environ = typing.MutableMapping[str, typing.Any]
StartResponse = typing.Callable[[str, typing.Iterable[typing.Tuple[str, str]]], None]
WSGIApp = typing.Callable[
    [Environ, StartResponse], typing.Iterable[typing.Union[str, bytes]]
]


def typeassert(func: typing.Callable) -> typing.Callable:
    """
    Force check parameter type by type hint

    * Parameters that use default values will not be checked
    * Parameters without type hint will not be checked

    """
    # If in optimized mode, disable type checking
    if not Config().DEBUG:
        return func

    @wraps(func)
    def wrapper(*args, **kwargs) -> typing.Any:
        # check params
        sig = signature(func)
        bound_values = sig.bind(*args, **kwargs)
        for name, value in bound_values.arguments.items():
            parameter = sig.parameters[name]

            if (
                parameter.annotation is parameter.empty
                or parameter.annotation == typing.Any
            ):
                continue
            if value is parameter.default:
                continue
            if not isinstance(value, parameter.annotation):
                raise TypeError(
                    f"Argument {name} must be {parameter.annotation} but got {type(value)}"
                )
        # check return
        result = func(*args, **kwargs)
        if "return" in func.__annotations__:
            return_type = func.__annotations__["return"]
            if result is return_type:
                pass
            elif not isinstance(result, return_type):
                raise TypeError(f"Return must be {return_type} but got {type(result)}")
        return result

    return wrapper


def typeasserts(*ty_args, **ty_kwargs) -> typing.Callable:
    """
    Type checking without parameter annotation

    * Cannot be used to check the return value

    """

    def decorate(func: typing.Callable) -> typing.Callable:
        # If in optimized mode, disable type checking
        if not Config().DEBUG:
            return func

        # Map function argument names to supplied types
        sig = signature(func)
        bound_types = sig.bind_partial(*ty_args, **ty_kwargs).arguments

        @wraps(func)
        def wrapper(*args, **kwargs) -> typing.Any:
            bound_values = sig.bind(*args, **kwargs)
            # Enforce type assertions across supplied arguments
            for name, value in bound_values.arguments.items():
                if name in bound_types:
                    if not isinstance(value, bound_types[name]):
                        raise TypeError(
                            f"Argument {name} must be {bound_types[name]} but got {type(value)}"
                        )
            return func(*args, **kwargs)

        return wrapper

    return decorate



import copy
import typing
import inspect
import functools


def currying(func: typing.Callable) -> typing.Callable:
    f = func

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> typing.Any:
        nonlocal f
        if args or kwargs:
            f = functools.partial(f, *args, **kwargs)
            return wrapper
        return f()

    return wrapper


def merge_mapping(x: typing.Mapping, default: typing.Mapping) -> typing.Mapping:
    """merge x to default. return a *new* Mapping"""
    default = copy.deepcopy(default)
    for key, value in x.items():
        if isinstance(default.get(key), typing.Mapping):
            if not isinstance(value, typing.Mapping):
                raise ValueError(
                    f"{key} in default is a mapping, but {key} in x is not mapping."
                )
            merge_mapping(value, default[key])
            continue
        default[key] = x[key]
    return default



import os
import copy
import typing
import asyncio
import traceback
import importlib
from inspect import signature

from starlette.types import Scope, Receive, Send, Message, ASGIApp
from starlette.requests import Request
from starlette.websockets import WebSocket, WebSocketClose
from starlette.responses import RedirectResponse
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.errors import ServerErrorMiddleware
from starlette.middleware.wsgi import WSGIMiddleware
from starlette.exceptions import HTTPException, ExceptionMiddleware

from .responses import FileResponse, automatic
from .types import WSGIApp
from .config import config


async def favicon(scope: Scope, receive: Receive, send: Send) -> None:
    """
    favicon.ico
    """
    if scope["type"] == "http" and os.path.exists(os.path.normpath("favicon.ico")):
        response = FileResponse("favicon.ico")
        await response(scope, receive, send)
        return
    raise HTTPException(404)


class Lifespan:
    def __init__(
        self,
        on_startup: typing.List[typing.Callable] = None,
        on_shutdown: typing.List[typing.Callable] = None,
    ) -> None:
        self.on_startup = [] if on_startup is None else list(on_startup)
        self.on_shutdown = [] if on_shutdown is None else list(on_shutdown)

    def on_event(self, event_type: str) -> typing.Callable:
        """Wrapper add_event_type"""

        def add_event_handler(func: typing.Callable) -> typing.Callable:
            self.add_event_handler(event_type, func)
            return func

        return add_event_handler

    def add_event_handler(self, event_type: str, func: typing.Callable) -> None:
        if event_type == "startup":
            self.on_startup.append(func)
        elif event_type == "shutdown":
            self.on_shutdown.append(func)
        else:
            raise ValueError("event_type must be in ('startup', 'shutdown')")

    async def startup(self) -> None:
        """
        Run any `.on_startup` event handlers.
        """
        for handler in self.on_startup:
            if asyncio.iscoroutinefunction(handler):
                await handler()
            else:
                handler()

    async def shutdown(self) -> None:
        """
        Run any `.on_shutdown` event handlers.
        """
        for handler in self.on_shutdown:
            if asyncio.iscoroutinefunction(handler):
                await handler()
            else:
                handler()

    async def lifespan(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Handle ASGI lifespan messages, which allows us to manage application
        startup and shutdown events.
        """
        message = await receive()
        assert message["type"] == "lifespan.startup"

        try:
            await self.startup()
        except BaseException:
            msg = traceback.format_exc()
            await send({"type": "lifespan.startup.failed", "message": msg})
            raise

        await send({"type": "lifespan.startup.complete"})
        message = await receive()
        assert message["type"] == "lifespan.shutdown"
        await self.shutdown()
        await send({"type": "lifespan.shutdown.complete"})

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert scope["type"] == "lifespan"
        await self.lifespan(scope, receive, send)


def get_pathlist(uri: str) -> typing.List[str]:
    if uri.endswith("/index"):
        return RedirectResponse(f'/{uri[:-len("/index")]}', status_code=301)

    if uri.endswith("/"):
        uri += "index"

    filepath = uri[1:].strip(".")
    # Google SEO
    if not config.ALLOW_UNDERLINE:
        if "_" in uri:
            return RedirectResponse(f'/{uri.replace("_", "-")}', status_code=301)
        filepath = filepath.replace("-", "_")

    # judge python file
    abspath = os.path.join(config.path, "views", filepath + ".py")
    if not os.path.exists(abspath):
        raise HTTPException(404)

    pathlist = filepath.split("/")
    pathlist.insert(0, "views")

    return pathlist


class Filepath:
    def __init__(self) -> None:
        self.apps = {}
        self.lifespan = Lifespan()

    def mount(self, route: str, app: typing.Union[ASGIApp, WSGIApp]) -> None:
        assert route.startswith("/"), "prefix must be start with '/'"
        assert not route.endswith("/"), "prefix can't end with '/'"
        self.apps.update({route: app})

    async def http(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = Request(scope, receive)
        pathlist = get_pathlist(request.url.path)
        # find http handler
        module_path = ".".join(pathlist)
        module = importlib.import_module(module_path)
        try:
            get_response = module.HTTP()
        except AttributeError:
            raise HTTPException(404)
        # call middleware
        for deep in range(len(pathlist), 0, -1):
            try:
                module = importlib.import_module(".".join(pathlist[:deep]))
                get_response = module.Middleware(get_response)
            except AttributeError:
                continue
        # get response
        response = await get_response(request)
        if not isinstance(response, tuple):
            response = (response,)
        response = automatic(*response)
        await response(scope, receive, send)

    async def websocket(self, scope: Scope, receive: Receive, send: Send) -> None:
        websocket = WebSocket(scope, receive=receive, send=send)
        try:
            pathlist = get_pathlist(websocket.url.path)
            # find websocket handler
            module_path = ".".join(pathlist)
            module = importlib.import_module(module_path)
            try:
                handler = module.Socket()
            except AttributeError:
                raise HTTPException(404)
        except HTTPException as exception:
            if exception.status_code == 404:
                websocket_close = WebSocketClose()
                await websocket_close(receive, send)
                return
            raise exception
        await handler(websocket)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert scope["type"] in ("http", "websocket", "lifespan")

        if scope["type"] == "lifespan":
            await self.lifespan(scope, receive, send)
            return

        async def subsend(message: Message) -> None:
            if message["type"] == "http.response.start" and message["status"] == 404:
                raise HTTPException(404)
            return await send(message)

        async def callapp(
            app: typing.Union[ASGIApp, WSGIApp],
            scope: Scope,
            receive: Receive,
            send: Send,
        ) -> None:
            sig = signature(app)
            if len(sig.parameters) == 3:
                await app(scope, receive, send)
                return
            elif len(sig.parameters) == 2:
                if scope["type"] != "http":
                    raise HTTPException(404)

                app = WSGIMiddleware(app)
                await app(scope, receive, send)
                return

        path = scope["path"]
        root_path = scope.get("root_path", "")

        # Call into a submounted app, if one exists.
        for path_prefix, app in self.apps.items():
            if path.startswith(path_prefix):
                subscope = copy.deepcopy(scope)
                subscope["path"] = path[len(path_prefix) :]
                subscope["root_path"] = root_path + path_prefix
                try:
                    await callapp(app, subscope, receive, subsend)
                    return
                except HTTPException as exception:
                    if exception.status_code != 404:
                        raise exception

        await getattr(self, scope["type"])(scope, receive, send)


class Index:
    def __init__(self, debug: bool = False) -> None:
        self._debug = debug
        self.router = Filepath()
        self.exception_middleware = ExceptionMiddleware(self.router, debug=debug)
        self.error_middleware = ServerErrorMiddleware(
            self.exception_middleware, debug=debug
        )

    @property
    def debug(self) -> bool:
        return self._debug

    @debug.setter
    def debug(self, value: bool) -> None:
        self._debug = value
        self.exception_middleware.debug = value
        self.error_middleware.debug = value

    def add_middleware(self, middleware_class: type, **kwargs: typing.Any) -> None:
        self.error_middleware.app = middleware_class(
            self.error_middleware.app, **kwargs
        )

    def add_exception_handler(
        self,
        exc_class_or_status_code: typing.Union[int, typing.Type[Exception]],
        handler: typing.Callable,
    ) -> None:
        if exc_class_or_status_code in (500, Exception):
            self.error_middleware.handler = handler
        else:
            self.exception_middleware.add_exception_handler(
                exc_class_or_status_code, handler
            )

    def add_event_handler(self, event_type: str, func: typing.Callable) -> None:
        self.router.lifespan.add_event_handler(event_type, func)

    def on_event(self, event_type: str) -> typing.Callable:
        return self.router.lifespan.on_event(event_type)

    def exception_handler(
        self, exc_class_or_status_code: typing.Union[int, typing.Type[Exception]]
    ) -> typing.Callable:
        def decorator(func: typing.Callable) -> typing.Callable:
            self.add_exception_handler(exc_class_or_status_code, func)
            return func

        return decorator

    def middleware(self, middleware_type: str) -> typing.Callable:
        assert (
            middleware_type == "http"
        ), 'Currently only middleware("http") is supported.'

        def decorator(func: typing.Callable) -> typing.Callable:
            self.add_middleware(BaseHTTPMiddleware, dispatch=func)
            return func

        return decorator

    def mount(self, route: str, app: typing.Union[ASGIApp, WSGIApp]) -> None:
        self.router.mount(route, app)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["app"] = self
        await self.error_middleware(scope, receive, send)



import json
import typing
from inspect import signature
from functools import wraps

from starlette.requests import Request

from .models import Model
from .utils import currying


class ParseError(Exception):
    def __init__(self, error: typing.Any) -> None:
        self.error = error


async def partial(
    handler: typing.Callable, request: Request
) -> typing.Optional[typing.Any]:
    handler = currying(handler)
    sig = signature(handler)
    query = sig.parameters.get("query")
    if query and issubclass(query.annotation, Model):
        _query = query.annotation(request.query_params)
        query_error = await _query.clean()
        if query_error:
            raise ParseError({"error": {"query": query_error}})
        handler = handler(query=_query)

    body = sig.parameters.get("body")
    if body and issubclass(body.annotation, Model):
        if body.annotation.get_content_type() == "application/json":
            try:
                _body_data = await request.json()
            except json.decoder.JSONDecodeError:
                raise ParseError(
                    {"error": {"body": "You must submit a valid JSON string."}}
                )
        else:
            _body_data = await request.form()
        _body = body.annotation(_body_data)
        body_error = await _body.clean()
        if body_error:
            raise ParseError({"error": {"body": body_error}})
        handler = handler(body=_body)
    return handler


def bindresponse(
    status: int, response_model: Model = None, description: str = ""
) -> typing.Callable:
    """bind status => response model in http handler"""

    def decorator(func: typing.Callable) -> typing.Callable:
        """bind response model"""
        if hasattr(func, "__resps__"):
            getattr(func, "__resps__")[status] = {"model": response_model}
        else:
            setattr(func, "__resps__", {status: {"model": response_model}})

        getattr(func, "__resps__")[status]["description"] = description

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    return decorator


class Schema:
    @staticmethod
    def parameters(model: Model, position: str) -> typing.Dict[str, typing.Any]:
        """
        create openapi parameters docs from model.

        enum: "path", "query", "header", "cookie"

        https://swagger.io/docs/specification/describing-parameters/
        """
        assert position in ("path", "query", "header", "cookie")

        result = []
        for name, field in model.fields.items():
            schema = field.openapi()
            description = schema.pop("description")
            result.append(
                {
                    "in": position,
                    "name": name,
                    "schema": schema,
                    "description": description,
                    "required": not field.allow_null,
                }
            )
        return result

    @staticmethod
    def in_path(model: Model) -> typing.Dict[str, typing.Any]:
        return Schema.parameters(model, "path")

    @staticmethod
    def in_query(model: Model) -> typing.Dict[str, typing.Any]:
        return Schema.parameters(model, "query")

    @staticmethod
    def in_header(model: Model) -> typing.Dict[str, typing.Any]:
        return Schema.parameters(model, "header")

    @staticmethod
    def in_cookie(model: Model) -> typing.Dict[str, typing.Any]:
        return Schema.parameters(model, "cookie")

    @staticmethod
    def request_body(model: Model) -> typing.Dict[str, typing.Any]:
        return {
            "type": "object",
            "required": [
                name for name, field in model.fields.items() if not field.allow_null
            ],
            "properties": {
                name: field.openapi() for name, field in model.fields.items()
            },
        }

    @staticmethod
    def response(model: Model) -> typing.Dict[str, typing.Any]:
        return {
            "type": "object",
            "properties": {
                name: field.openapi() for name, field in model.fields.items()
            },
        }
